import base64
import cv2
import io
import numpy as np
import re
from datetime import datetime
from PIL import Image
from bankstatementextractor_final.banks_utils import *
import json
import fitz  # PyMuPDF library
import PyPDF2  # PyPDF2 library
import os
import subprocess
import torch
from pdf2image import convert_from_path
from unidecode import unidecode
import pandas as pd
import itertools    
os.sys.path
from io import StringIO
import traceback 

class Banks:

    def __init__(self):
        pass
    # adcb_1
    def adcb_1(self,pdf_bytes):
        try:
            doc = fitz.open(stream=pdf_bytes, filetype="pdf")
            output = []  # Initialize the output list

            for page in doc:
                try:
                    page_output = page.get_text("blocks")
                    output.append(page_output)
                except Exception as e:
                    print(f"Error while processing page: {e}")
                    return None

            plain_text_data = []

            for page_output in output:
                previous_block_id = 0
                page_plain_text_data = []

                for block in page_output:
                    try:
                        if block[6] == 0:
                            if previous_block_id != block[5]:
                                plain_text = unidecode(block[4])
                                page_plain_text_data.append(plain_text)
                                previous_block_id = block[5]
                    except Exception as e:
                        print(f"Error while processing block: {e}")
                        return None

                if 'Consolidated Statement\n' in page_plain_text_data:
                    continue

                page_plain_text_data = [text for text in page_plain_text_data if not text.startswith(('balance', 'opening'))]

                plain_text_data.append(page_plain_text_data)

            account_number = None
            name = None
            opening_balance = None
            closing_balance = None
            currency_code = None

            # Iterate through the first sublist
            for line in plain_text_data[0]:
                try:
                    if line.startswith('Account Name(s)'):
                        name = line.split('\n')[1]
                    elif line.startswith('Account Number'):
                        account_number_with_currency = line.split('\n')[1]
                        account_number = account_number_with_currency.split(' ')[0]
                        currency_code = account_number_with_currency.split(' ')[1]   
                    elif line.startswith('Opening Balance'):
                        opening_balance_with_currency = line.split('\n')[1]
                        opening_balance = float(opening_balance_with_currency.replace(',', '').replace('AED', ''))
                    elif line.startswith('Closing Balance'):
                        closing_balance_with_currency = line.split('\n')[1]
                        closing_balance = float(closing_balance_with_currency.replace(',', '').replace('AED', ''))
                except Exception as e:
                    print(f"Error while processing account info: {e}")
                    return None

            obj = {
                "account_id": "",
                "name": name,
                "currency_code": currency_code,
                "type": "",
                "iban": "",
                "account_number": account_number,
                "bank_name": "ADCB",
                "branch": "",
                "credit": "",
                "address":""
            }       
            new_lst = []

            for sublist in plain_text_data:
                keep = False  # Flag to indicate whether to keep the elements

                for i, element in enumerate(sublist):
                    if element.startswith('Posting Date\nValue Date\nDescription\nRef/Cheque No'):
                        keep = True  
                        new_sublist = sublist[i+1:]  # Create a new sublist starting from the target element
                        break

                if keep:
                    new_lst.append(new_sublist)  # Add the new sublist to the result list if the target element is found


            flat_data = [item for sublist in new_lst for item in sublist]
            split_data = [entry.strip().split('\n') for entry in flat_data]

            lst = []
            current_element = []

            pattern = re.compile(r'^\d{2}/\d{2}/\d{4}')

            for sublist in split_data:
                if pattern.match(sublist[0]):
                    if current_element:
                        lst.append(current_element)
                    current_element = sublist
                else:
                    current_element += sublist

            if current_element:
                lst.append(current_element)

            lst_1 = []

            for sublist in lst:
                if len(sublist) > 6:
                    new_sublist = sublist[:2] + sublist[-3:]
                    new_sublist.insert(2, ' '.join(sublist[2:-3]))
                    lst_1.append(new_sublist)
                else:
                    lst_1.append(sublist)

            # Create a DataFrame with specific column names
            df = pd.DataFrame(lst_1, columns=['timestamp', 'Value Date', 'description',  'debit', 'credit', 'running_balance'])
            df['debit'] = df['debit'].str.replace(',', '').astype(float)
            df['credit'] = df['credit'].str.replace(',', '').astype(float)
            df['timestamp'] = pd.to_datetime(df['timestamp'], format='%d/%m/%Y')
            df['running_balance'] = df['running_balance'].str.replace(',', '').astype(float)
            df['amount'] = df.apply(lambda row: -row['debit'] if row['debit'] != 0 else row['credit'], axis=1)
            for key, value in obj.items():
                df[key] = value
            df.drop(['debit', 'credit','Value Date'], axis=1, inplace = True)
            error_flags = {
                "data_cover_last_3_months": False,
                "statement_extracted_last_7_days": False,
                "outstanding_amount_match": False
            }

            # # Check if the data covers the last 3 months
            # if (df['timestamp'].max() - df['timestamp'].min()).days < 90:
            #     error_flags["data_cover_last_3_months"] = True
            #     print("Error: Upload a 3 months bank statement")
            #     return None

            # # Check if the statement was extracted within the last 7 days of upload
            # if (datetime.now().date() - df['timestamp'].max().to_pydatetime().date()).days > 7:
            #     error_flags["statement_extracted_last_7_days"] = True
            #     print("Error: Upload a statement extracted within the last 7 days")
            #     return None

            # #Check if the outstanding amount matches the expected total
            # if round(abs(df['amount'].sum()), 2) != round(abs(opening_balance - closing_balance), 2):
            #     error_flags["outstanding_amount_match"] = True
            #     print("Error: Upload not edited bank statement") 
            #     None
            
            df['timestamp'] = df['timestamp'].dt.strftime('%Y-%m-%d')
            json_data = df.to_json(orient='records')
            return json_data
        
        except Exception as e:
            print(f"Error during processing: {e}")
            traceback.print_exc()
            return None

    def liv_1(self,pdf_bytes):
        try:
            doc = fitz.open(stream=pdf_bytes, filetype="pdf")
            output = []
            for page in doc:
                page_output = page.get_text("blocks")  # Get the text blocks for each page
                output.append(page_output)  # Append the page output to the main output list

            plain_text_data = []  # Initialize an empty list to store the plain text

            for page_output in output:
                previous_block_id = 0  # Set a variable to mark the block id
                page_plain_text_data = []  # Initialize an empty list to store the plain text for each page

                for block in page_output:
                    if block[6] == 0:  # We only take the text
                        if previous_block_id != block[5]:  # Compare the block number
                            plain_text = unidecode(block[4])
                            page_plain_text_data.append(plain_text)  # Store the plain text in the list for the current page
                            previous_block_id = block[5]  # Update the previous block id

                # Check if 'Consolidated Statement' is present in the sublist and delete the sublist
                if 'Consolidated Statement\n' in page_plain_text_data:
                    continue  # Skip the current page 
                plain_text_data.append(page_plain_text_data)

            # Initialize variables to store information
            name = None
            address = None
            statement_period = None
            account_balance = None
            account_no = None
            iban = None

            # Iterate through the list and search for the respective information
            for item in plain_text_data[0]:
                if item.startswith('Name :'):
                    name = item.split('Name :')[1].replace('\n', '')
                elif item.startswith('Address :'):
                    address = item.split('Address :')[1].replace('\n', '')
                elif item.startswith('Statement period :'):
                    statement_period = item.split('Statement period :')[1].replace('\n', '')
                elif item.startswith('Account balance :'):
                    account_balance = item.split('Account balance :')[1].replace('\n', '')
                elif item.startswith('AccountNo :'):
                    account_no = item.split('AccountNo :')[1].replace('\n', '')
                elif item.startswith('IBAN :'):
                    iban = item.split('IBAN :')[1].replace('\n', '')

            obj = {
                'account_id':'',
                'name': name,
                'currency_code': 'AED',
                'type': '',
                'iban': iban,
                'account_number': account_no,
                'bank_name': 'LIV by ENBD',
                'branch': '',
                'address': address,
                'account_balance': account_balance
            }

            opening_balance_index = next((i for sublist in plain_text_data for i, text in enumerate(sublist) if text.startswith('IBAN :')), -1)

            if opening_balance_index != -1:
                for i in range(len(plain_text_data)):
                # Remove all elements before opening_balance_index + 1 in each sublist
                    plain_text_data[i] = plain_text_data[i][opening_balance_index + 1:]


            plain_text_data = [[element for element in sublist if not element.startswith(('Confirmation of the correctness', 'This statement is generated on', 'Money in'))] for sublist in plain_text_data]
            plain_text_data = list(itertools.chain.from_iterable(plain_text_data))
            plain_text_data = [element.rstrip() for element in plain_text_data]
            plain_text_data = [elem.split('\n') for elem in plain_text_data]

            # Initialize empty lists to store data
            dates = []
            transactions = []
            amounts = []
            balances = []

            # Regex pattern to match date'dd/mm/yyyy'
            date_pattern = r'\d{2}/\d{2}/\d{4}'

            # Initialize a transaction variable to concatenate transaction details
            transaction = ""

            # Iterate through the list
            for item in plain_text_data:
                if re.match(date_pattern, item[0]):
                    # If the item is a date, store it and reset the transaction variable
                    date = item[0]
                    transaction = ""
                else:
                    # Concatenate the elements in the sublist after the date to form the transaction
                    transaction += " ".join(item) + " "

                    # Check if the item contains an amount and balance
                    amount_match = re.search(r'([-+]?\d{1,3}(?:,\d{3})*\.\d+)', item[0])
                    if amount_match:
                        amount = amount_match.group(1).replace(',', '')  # Remove commas
                        balance = item[-1].replace(',', '')  # Remove commas
                        # Append data to respective lists
                        dates.append(pd.to_datetime(date, format='%d/%m/%Y'))
                        transactions.append(transaction.strip())  # Remove trailing spaces
                        amounts.append(pd.to_numeric(amount))
                        balances.append(pd.to_numeric(balance))

            # Create a DataFrame
            df = pd.DataFrame({'timestamp': dates, 'description': transactions, 'amount': amounts, 'running_balance': balances})
            
            
            for key, value in obj.items():
                df[key] = value
            error_flags = {
                "data_cover_last_3_months": False,
                "statement_extracted_last_7_days": False,
                "outstanding_amount_match": False
            }

            # # Check if the data covers the last 3 months
            # if (df['timestamp'].max() - df['timestamp'].min()).days < 90:
            #     error_flags["data_cover_last_3_months"] = True
            #     print("Error: Upload a 3 months bank statement")
            #     return None

            # # Check if the statement was extracted within the last 7 days of upload
            # if (datetime.now().date() - df['timestamp'].max().to_pydatetime().date()).days > 7:
            #     error_flags["statement_extracted_last_7_days"] = True
            #     print("Error: Upload a statement extracted within the last 7 days")
            #     return None

    #         #Check if the outstanding amount matches the expected total
    #         if round(abs(df['amount'].sum()), 2) != round(abs(opening_balance - closing_balance), 2):
    #             error_flags["outstanding_amount_match"] = True
    #             print("Error: Upload not edited bank statement")
                    
            df['timestamp'] = df['timestamp'].dt.strftime('%Y-%m-%d')
            json_data = df.to_json(orient='records')
            return json_data

        except Exception as e:
            print(f"Error during processing: {e}")
            traceback.print_exc()
            return None
    def enbd_1(self,pdf_bytes):
        try:
            doc = fitz.open(stream=pdf_bytes, filetype="pdf")
            plain_text_data = []

            for page in doc:
                page_output = [unidecode(block[4]) for block in page.get_text("blocks") if block[6] == 0]
                plain_text_data.append(page_output)

            # Personal details extraction
            elements_above_target = []

            # Iterate through the elements in the first sublist
            for item in plain_text_data[0]:
                if item.startswith('Date\nDescription\n'):
                    break
                elements_above_target.append(item)
            elements_above_target = [element.replace('\n', ' ') for element in elements_above_target]

            flattened_string = ' '.join(elements_above_target)

            name = None
            branch = None
            opening_balance = None
            account_type = None
            iban = None
            currency_info = None
            account_number = None

            # Check if the flattened string contains the required details
            if 'Name' in flattened_string and 'Branch' in flattened_string and 'Opening Balance' in flattened_string and 'Account Type' in flattened_string and '(IBAN)' in flattened_string and 'Currency' in flattened_string:
                name = flattened_string.split('Name')[1].split('Branch')[0].strip()
                branch = flattened_string.split('Branch')[1].split('Opening Balance')[0].strip()
                opening_balance = flattened_string.split('Opening Balance')[1].strip()
                currency_info = flattened_string.split('Currency')[1].split('Account Number')[0].strip()
                account_type = flattened_string.split('Account Type')[1].split('Name')[0].strip()
                temp = flattened_string.split('(IBAN)')[1].split('Account Type')[0].strip()
                open_bracket_index = temp.find("(")
                close_bracket_index = temp.find(")")
                iban = temp[open_bracket_index + 1:close_bracket_index]  
                account_number = temp[:open_bracket_index]  

            obj = {
                "account_id": "",
                "name": name,
                "currency_code": currency_info,
                "type": account_type,
                "iban": iban,
                "account_number": account_number,
                "bank_name": "ENBD",
                "branch": branch,
                #"credit": "",
                "address": ""
            }

            # Transaction Extraction
            for i, sublist in enumerate(plain_text_data):
                for j, item in enumerate(sublist):
                    if item.startswith('Date\nDescription\nDebit\nCredit\nAccount Balance\n'):
                        # Found the starting element, remove all elements before it
                        plain_text_data[i] = sublist[j + 1:]
                        break

            flat_data = [item for sublist in plain_text_data for item in sublist]
            split_data = [entry.strip().split('\n') for entry in flat_data]

            date_pattern = r'\d{2} \w{3} \d{4}' 


            matching_indexes = []
            for idx, sublist in enumerate(split_data):
                if len(sublist) > 0 and re.match(date_pattern, sublist[0]):
                    matching_indexes.append(idx)
            data = []
            amount = ''
            running_balance = ''
            # Iterate through the indexes of matching sublists
            for idx in matching_indexes:
                try:
                    sublist = split_data[idx]

                    # Extract the date value from the first element that matches the date pattern
                    timestamp = re.search(date_pattern, sublist[0]).group()
                    date_value = pd.to_datetime(timestamp, format='%d %b %Y')

                    # Join the remaining elements in the sublist to form the description
                    description = ' '.join(sublist[1:])

                    # Check if there's a sublist after the current one
                    if idx + 1 < len(split_data):
                        next_sublist = split_data[idx + 1]

                        # Check if the first element of the next sublist matches the amount format
                        if re.match(r'[-]?\d{1,3}(,\d{3})*\.\d{2}', next_sublist[0]):
                            # If yes, add it to the 'amount' column
                            amount = next_sublist[0]

                            # Check if the last element of the next sublist matches the running balance format
                            if re.match(r'AED [-]?\d{1,3}(,\d{3})*\.\d{2}', next_sublist[-1]):
                                # If yes, extract it as 'running_balance'
                                running_balance = next_sublist[-1]

                    # Append the data to the list
                    data.append([timestamp, description, amount, running_balance])
                except Exception as e:
                    print(f"Error while processing transaction data: {e}")

            # Create a DataFrame from the processed data
            df = pd.DataFrame(data, columns=['timestamp', 'description', 'amount', 'running_balance'])
            df['amount'] = df['amount'].str.replace(',', '').astype(float)
            df['running_balance'] = df['running_balance'].str.replace('AED ', '').str.replace(',', '').astype(float)
            df['timestamp'] = pd.to_datetime(df['timestamp'], format='%d %b %Y', errors='coerce')
            for key, value in obj.items():
                df[key] = value
            error_flags = {
                "data_cover_last_3_months": False,
                "statement_extracted_last_7_days": False,
                "outstanding_amount_match": False
            }

            # # Check if the data covers the last 3 months
            # if (df['timestamp'].max() - df['timestamp'].min()).days < 90:
            #     error_flags["data_cover_last_3_months"] = True
            #     print("Error: Upload a 3 months bank statement")

            # # Check if the statement was extracted within the last 7 days of upload
            # if (datetime.now().date() - df['timestamp'].max().to_pydatetime().date()).days > 7:
            #     error_flags["statement_extracted_last_7_days"] = True
            #     print("Error: Upload a statement extracted within the last 7 days")

            df['timestamp'] = df['timestamp'].dt.strftime('%Y-%m-%d')
            
            json_data = df.to_json(orient='records')
            return json_data

        except Exception as e:
            print(f"Error while processing the PDF: {e}")
            return None
    def adib_1(self, pdf_bytes):
        try:
            doc = fitz.open(stream=pdf_bytes, filetype="pdf")
            output = []

            for page in doc:
                page_output = page.get_text("blocks")
                output.append(page_output)

            plain_text_data = []

            for page_output in output:
                previous_block_id = 0
                page_plain_text_data = []

                for block in page_output:
                    if block[6] == 0:
                        if previous_block_id != block[5]:
                            plain_text = unidecode(block[4])
                            page_plain_text_data.append(plain_text)
                            previous_block_id = block[5]

                if 'Consolidated Statement\n' in page_plain_text_data:
                    continue

                page_plain_text_data = [text for text in page_plain_text_data if not text.startswith(('balance', 'opening'))]

                plain_text_data.append(page_plain_text_data)

            name = None
            address = None
            account_number = None
            acc_type = None
            currency_code = None

            for i, element in enumerate(plain_text_data[0]):
                if element.startswith('Account'):
                    name = plain_text_data[0][i - 2].strip()
                    address = plain_text_data[0][i - 1].strip()
                    account_info = plain_text_data[0][i + 1].strip().split('\n')
                    account_number = account_info[0].strip()
                    acc_type = account_info[1].strip()
                    currency_code = account_info[2].strip()
                else:
                    name = plain_text_data[0][0].split('\n')[0]
                    address = ' '.join(plain_text_data[0][0].split('\n')[1:])
                    account_number = plain_text_data[0][2].split('\n')[0]
                    acc_type = plain_text_data[0][2].split('\n')[1]
                    currency_code =plain_text_data[0][2].split('\n')[2]

            obj = {
                "account_id": "",
                "name": name,
                "currency_code": currency_code,
                "type": acc_type,
                "iban": "",
                "account_number": account_number,
                "bank_name": "ADIB",
                "branch": "",
                #"credit": "",
                "address": address
            }

            flat_data = [item for sublist in plain_text_data for item in sublist]

            ob = r'Opening Balance\n([\d,.]+)'
            opening_balance = None

            for item in flat_data:
                match = re.search(ob, item)
                if match:
                    opening_balance = match.group(1)
                    opening_balance = float(opening_balance.replace(',', ''))
                    break

            cb = r'Closing Balance \n([\d,.]+)'
            closing_balance = None

            for item in flat_data:
                match = re.search(cb, item)
                if match:
                    closing_balance = match.group(1)
                    closing_balance = float(closing_balance.replace(',', ''))
                    break

            pattern = r'(\d{2}-\d{2}-\d{4})\s+([\s\S]*?)\n([\d,]+\.\d{2})\n([\d,]+\.\d{2})'
            matches = re.findall(pattern, '\n'.join(flat_data))
            data_list = []
            previous_amount2 = None

            for match in matches:
                date = match[0]
                description = match[1]
                amount1 = match[2]
                amount2 = match[3]
                amount2_numeric = float(amount2.replace(',', ''))

                if previous_amount2 is not None and amount2_numeric < previous_amount2:
                    amount1 = '-' + amount1

                previous_amount2 = amount2_numeric

                data_list.append((date, description, amount1, amount2))

            columns = ["timestamp", "description", "amount", "running_balance"]
            df = pd.DataFrame(data_list, columns=columns)
            df['amount'] = df['amount'].apply(lambda x: float(x.replace(',', '')))
            df['running_balance'] = df['running_balance'].apply(lambda x: float(x.replace(',', '')))
            df['timestamp'] = pd.to_datetime(df['timestamp'], format="%d-%m-%Y")

            if df.loc[0, 'running_balance'] < opening_balance:
                df.loc[0, 'amount'] = -df.loc[0, 'amount']

            for key, value in obj.items():
                df[key] = value

            error_flags = {
                "data_cover_last_3_months": False,
                "statement_extracted_last_7_days": False,
                "outstanding_amount_match": False
            }

            if (df['timestamp'].max() - df['timestamp'].min()).days < 90:
                error_flags["data_cover_last_3_months"] = True

            if (datetime.now().date() - df['timestamp'].max().to_pydatetime().date()).days > 7:
                error_flags["statement_extracted_last_7_days"] = True

            if round(abs(df['amount'].sum()), 2) != round(abs(opening_balance - closing_balance), 2):
                error_flags["outstanding_amount_match"] = True

            df['timestamp'] = pd.to_datetime(df['timestamp'], format='%d-%m-%Y')
            print(df)

            json_data = df.to_json(orient='records')
            print(obj)
            
            return json_data

        except Exception as e:
            print(f"Error: {str(e)}")
            return None

    def cbd_1(self, pdf_bytes):
        try:
            doc = fitz.open(stream=pdf_bytes, filetype="pdf")
            output = []
            data_ = []
            prev_balance = 0
            pattern = r'^\d{2}/\d{2}/\d{4}'  # Pattern to check if the transaction starts with a date
            count = 0
            name = ''
            currency = ''
            acc_type = ''
            iban = ''
            account_number = ''
            branch = '' 
            
            for page in doc:
                output += page.get_text("blocks")

            for tup in output:
                trans = tup[4].encode("ascii", errors="ignore").decode()

                # TODO: extracting account details
                if count == 2:
                    if trans.startswith('::\n'):
                        branch = output[count+1][4].replace('\n',',').rstrip(',')
                        name = output[count+2][4].strip().split('\n')[1].strip()
                    else:
                        branch = trans.replace('\n',',').rstrip(',')
                        name = output[count+1][4].strip().split('\n')[1].strip()

                if trans.lower().startswith('acct. no.') and account_number == '':
                    account_number = trans.strip().split('\n')[1].strip()

                if trans.lower().startswith('iban') and iban == '':
                    iban = trans.strip().split('\n')[1].strip()
                    acc_type = trans.strip().split('\n')[-1].strip()

                if trans.strip().lower().startswith('currency') and currency == '':
                    currency = trans.strip().split('\n')[1]
                    if "-" in currency:
                        currency = currency[currency.index('-')+1:].strip()

                if trans.strip().startswith('Balance Brought FWD'):
                    prev_balance = float(trans.split('\n')[1].replace(',', ''))

                # TODO: extracting transaction details
                if re.match(pattern, trans):
                    trans_list = trans.split('\n')
                    t_date = trans_list[0]

                    # TODO: extraction of description - number of lines = 1
                    if len(trans_list) == 6:
                        description = trans_list[1]
                        cur_balance = float(trans_list[4].replace(',', ''))

                        # check if the transaction is debit or credit
                        if float(cur_balance) > float(prev_balance):
                            amount = trans_list[3]
                        else:
                            amount = '-' + trans_list[3]

                    # TODO: extraction of description - number of lines = 2
                    elif len(trans_list) == 7:
                        description = trans_list[1] + ' ' + trans_list[2]
                        cur_balance = float(trans_list[5].replace(',', ''))

                        # check if the transaction is debit or credit
                        if float(cur_balance) > float(prev_balance):
                            amount = trans_list[4]
                        else:
                            amount = '-' + trans_list[4]

                    # TODO: extraction of description - number of lines = 3
                    elif len(trans_list) == 8:
                        description = trans_list[1] + ' ' + trans_list[2] + ' ' + trans_list[3] 
                        cur_balance = float(trans_list[6].replace(',', ''))

                        # check if the transaction is debit or credit
                        if float(cur_balance) > float(prev_balance):
                            amount = trans_list[5]
                        else:
                            amount = '-' + trans_list[5]

                    # TODO: extraction of description - number of lines = 4
                    elif len(trans_list) == 9:
                        description = trans_list[1] + ' ' + trans_list[2] + ' ' + trans_list[3] + ' ' + trans_list[4]
                        cur_balance = float(trans_list[7].replace(',', ''))

                        # check if the transaction is debit or credit
                        if float(cur_balance) > float(prev_balance):
                            amount = trans_list[6]
                        else:
                            amount = '-' + trans_list[6]

                    # TODO: extraction of description - number of lines = 5
                    elif len(trans_list) == 10:
                        description = trans_list[1] + ' ' + trans_list[2] + ' ' + trans_list[3] + ' ' + trans_list[4] + ' ' + trans_list[5]
                        cur_balance = float(trans_list[8].replace(',', ''))

                        # check if the transaction is debit or credit
                        if float(cur_balance) > float(prev_balance):
                            amount = trans_list[7]
                        else:
                            amount = '-' + trans_list[7]

                    obj = {
                            'name' : name,
                            'currency_code' : currency,
                            'type' : acc_type,
                            'iban' : iban,
                            'account_number' : account_number,
                            'bank_name' : 'Commercial Bank of Dubai',
                            'branch' : branch,
                            "timestamp": pd.to_datetime(t_date, format='%d/%m/%Y'),
                            "description":description,
                            "amount":amount,
                            "running_balance":cur_balance
                        }
                    data_.append(obj)
                    prev_balance = cur_balance

                count += 1

            df = pd.DataFrame(data_)
            error_flags = {
                    "data_cover_last_3_months": False,
                    "statement_extracted_last_7_days": False,
                    "outstanding_amount_match": False
                }

            # # Check if the data covers the last 3 months
            # if (df['timestamp'].max() - df['timestamp'].min()).days < 90:
            #     error_flags["data_cover_last_3_months"] = True
            #     print("Error: Upload a 3 months bank statement")
            #     return None

            # # Check if the statement was extracted within the last 7 days of upload
            # if (datetime.now().date() - df['timestamp'].max().to_pydatetime().date()).days > 7:
            #     error_flags["statement_extracted_last_7_days"] = True
            #     print("Error: Upload a statement extracted within the last 7 days")
            #     return None
            
            df['timestamp'] = df['timestamp'].dt.strftime('%Y-%m-%d')
            json_data = df.to_json(orient='records')
            print(json_data)
            return json_data

        except Exception as e:
            print(f"Error during processing: {e}")
            return None
    def adcb_2(self, pdf_bytes):
        try:
            doc = fitz.open(stream=pdf_bytes, filetype="pdf")
            output = []

            for page in doc:
                page_output = page.get_text("blocks")
                output.append(page_output)

            plain_text_data = []

            for page_output in output:
                previous_block_id = 0
                page_plain_text_data = []

                for block in page_output:
                    if block[6] == 0:
                        if previous_block_id != block[5]:
                            plain_text = unidecode(block[4])
                            page_plain_text_data.append(plain_text)
                            previous_block_id = block[5]
                plain_text_data.append(page_plain_text_data)

            iban_pattern = r'IBAN: (AE\d{20})'
            currency_pattern = r'Currency: ([A-Z]{3})'
            branch_pattern = r'Branch: (.*?)\nAccount Title:'
            name_pattern = r'Account Title: (.*)'
            account_details_pattern = r'Account Details: (.*?)\nCurrency:'
            account_merge_split_pattern = r' - '
            obj = {  # Initialize obj here to avoid "referenced before assignment" error
                "account_id": "",
                "name": "",
                "currency_code": "",
                "type": "",
                "iban": "",
                "account_number": "",
                "bank_name": "ADCB",
                "branch": "",
                "address": ""
        }
            for sublist in plain_text_data:
                iban = None
                currency = None
                branch = None
                name = None
                account_merge = None
                account_number = None
                acc_type = None

                for line in sublist:
                    iban_match = re.search(iban_pattern, line)
                    currency_match = re.search(currency_pattern, line)
                    branch_match = re.search(branch_pattern, line)
                    name_match = re.search(name_pattern, line)
                    account_details_match = re.search(account_details_pattern, line)

                    if iban_match:
                        iban = iban_match.group(1)

                    if currency_match:
                        currency = currency_match.group(1)

                    if branch_match:
                        branch = branch_match.group(1)

                    if name_match:
                        name = name_match.group(1)

                    if account_details_match:
                        account_merge = account_details_match.group(1)
                        account_parts = account_merge.split(account_merge_split_pattern)
                        if len(account_parts) == 2:
                            account_number, acc_type = account_parts
                            
                        obj = {
                            "account_id": "",
                            "name": name,
                            "currency_code": currency,
                            "type": acc_type,
                            "iban": iban,
                            "account_number": account_number,
                            "bank_name": "ADCB",
                            "branch": branch,
                            #"credit": "",
                            "address": ""
                        }     

                    if iban and currency and branch and name and account_number and acc_type:
                        break       

            filtered_list = [sublist for sublist in plain_text_data if not any(item.startswith('Consolidated Statement of Accounts') for item in sublist)]
            filtered_list_1 = [[entry for entry in sublist if not entry.startswith(('Total', 'End Of Statement'))] for sublist in filtered_list]
            new_lst = []

            for sublist in filtered_list_1:
                keep = False

                for i, element in enumerate(sublist):
                    if element.startswith('Date\nDescription\nChq/Ref No.'):
                        keep = True
                        new_sublist = sublist[i+2:]
                        break

                if keep:
                    new_lst.append(new_sublist)
            
            flat_data = [item for sublist in new_lst for item in sublist] 
            split_data = [entry.strip().split('\n') for entry in flat_data]

            data = []
            timestamp = None
            value_date = None
            description = None
            ref_cheque_no = None
            running_balance = None
            amount = None

            for sublist in split_data:
                if len(sublist) > 0 and re.match(r'\d{2}/\d{2}/\d{4}', sublist[0]):
                    timestamp = sublist[0]

                if len(sublist) >= 3 and re.match(r'\d{2}/\d{2}/\d{4}', sublist[-3]):
                    value_date = sublist[-3]

                if re.match(r'[-]?\d{1,3}(,\d{3})*\.\d{2}', sublist[-1]):
                    running_balance = sublist[-1]

                if len(sublist) >= 2 and re.match(r'[-]?\d{1,3}(,\d{3})*\.\d{2}', sublist[-2]):
                    amount = sublist[-2]
                description = ' '.join(sublist[1:-3])

                row_data = {
                    'timestamp': timestamp,
                    'Value Date': value_date,
                    'description': description,
                    'Ref/Cheque No': ref_cheque_no,
                    'running_balance': running_balance,
                    'amount': amount
                }

                data.append(row_data)

            df = pd.DataFrame(data, columns=['timestamp', 'description', 'amount', 'running_balance'])
            df['timestamp'] = pd.to_datetime(df['timestamp'], format='%d/%m/%Y')
            df['running_balance'] = df['running_balance'].str.replace(',', '').astype(float)
            df['amount'] = df['amount'].str.replace(',', '').astype(float)
            for i in range(1, len(df)):
                if df.at[i, 'running_balance'] < df.at[i - 1, 'running_balance']:
                    df.at[i, 'amount'] = -df.at[i, 'amount']
            for key, value in obj.items():
                df[key] = value
            
            error_flags = {
                "data_cover_last_3_months": False,
                "statement_extracted_last_7_days": False,
                "outstanding_amount_match": False
            }

            # if (df['timestamp'].max() - df['timestamp'].min()).days < 90:
            #     error_flags["data_cover_last_3_months"] = True
            #     print("Error: Upload a 3 months bank statement")

            # if (datetime.now().date() - df['timestamp'].max().to_pydatetime().date()).days > 7:
            #     error_flags["statement_extracted_last_7_days"] = True
            #     print("Error: Upload a statement extracted within the last 7 days")

            df['timestamp'] = df['timestamp'].dt.strftime('%Y-%m-%d')
            json_data = df.to_json(orient='records')

            return json_data

        except Exception as e:
            print(f"An error occurred: {str(e)}")
            return None
    
    def mashreq_2(self,pdf_bytes):
        try:
            doc = fitz.open(stream=pdf_bytes, filetype="pdf")
            plain_text_data = []

            # Extract text from the PDF
            for page in doc:
                page_output = [unidecode(block[4]) for block in page.get_text("blocks") if block[6] == 0]
                plain_text_data.append(page_output)

            # Filter out unwanted lines
            filtered_list = []
            for sublist in plain_text_data:
                filtered_sublist = [item for item in sublist if not item.startswith(('This is a system generated', 'Page', 'Closing Balance', 'Opening Balance'))]
                filtered_list.append(filtered_sublist)

            # Regular expressions for account number and currency code
            acc_pattern = r'\b\d{12}\n'
            cur_pattern = r'\b[A-Z]{3}\n'

            # Extract account information
            name = (filtered_list[0][1] if filtered_list[0][0].startswith('Account Statement') else filtered_list[0][0]).rstrip()
            account_number = next((re.match(acc_pattern, element).group(0).rstrip() for element in filtered_list[0] if re.match(acc_pattern, element)), None)
            account_type = (filtered_list[0][filtered_list[0].index(account_number+'\n') + 1]).rstrip()
            currency = next((re.match(cur_pattern, element).group(0).rstrip() for element in filtered_list[0] if re.match(cur_pattern, element)), None)

            # Extract closing balances
            opening_balance = None
            closing_balance = None
            for element in filtered_list[0]:
                if re.match(r'\d{1,3}(,\d{3})*\.\d+\n', element):
                    if opening_balance is None:
                        opening_balance = float(element.rstrip().replace(',', ''))
                    else:
                        closing_balance = float(element.rstrip().replace(',', ''))

            # Create the final object
            obj = {
                "account_id": "",
                "name": name,
                "currency_code": currency,
                "type": account_type,
                "iban": "",
                "account_number": account_number,
                "bank_name": "Mashreq",
                "branch": "",
                "credit": "",
                "address": ""
            }

            # Extract and process the transaction data
            new_lst = []
            for sublist in filtered_list:
                for i, element in enumerate(sublist):
                    if element.startswith('Balance\n'):
                        new_lst.append(sublist[i+2:])
                        break

            # Flatten and split the transaction data
            flattened_lst = [item.strip('\n').split('\n') for sublist in new_lst for item in sublist]

            data = []

            for sublist in flattened_lst:
                if len(sublist) > 5:
                    row_data = {
                        'timestamp': sublist[0],
            #             'value_date': sublist[1],
            #             'reference': sublist[2],
                        'description': ' '.join(sublist[3:-3]),
                        'running_balance': float(sublist[-1].replace(',', '')),
                        'amount': sublist[-2]
                    }
                    data.append(row_data)
                else:
                    # Initialize a list to store the extracted sublists
                    lst_1 = []
                    date_pattern = r'\d{2} \w{3} \d{4}'
                    current_sublist = []

                    # Iterate through the flattened list
                    for sublist in flattened_lst:
                        if re.match(date_pattern, sublist[0]):
                            if current_sublist:
                                lst_1.append(current_sublist)
                            current_sublist = sublist
                        else:
                            current_sublist.extend(sublist)

                    # Append the last sublist
                    if current_sublist:
                        lst_1.append(current_sublist)

                    # Process the data in lst_1 and update the 'data' list
                    for sublist in lst_1:
                        if len(sublist) > 5:
                            data.append({
                                'timestamp': sublist[0],
            #                     'value_date': sublist[1],
            #                     'reference': sublist[2],
                                'description': ' '.join(sublist[3:-3]),
                                'running_balance': float(sublist[-1].replace(',', '')),
                                'amount': sublist[-2]
                            })     
            df = pd.DataFrame(data)
            df['amount'] = df['amount'].apply(lambda x: float(x.replace(',', '').replace('+', '')))
            df['timestamp'] = pd.to_datetime(df['timestamp'], format='%d %b %Y', errors='coerce')
            json_data = df.to_json(orient='records')
            return json_data
        except Exception as e:
            print(f"Error during processing: {e}")
            return None   
    
    def ei_1(self, pdf_bytes):
        try:
            doc = fitz.open(stream=pdf_bytes, filetype="pdf")
            plain_text_data = []

            for page in doc:
                page_output = [unidecode(block[4]) for block in page.get_text("blocks") if block[6] == 0]
                plain_text_data.append(page_output)

            account_number = None
            account_type = None

            for element in plain_text_data[0]:
                if element.startswith('Account Number'):
                    # Extract the 13-digit number
                    account_number_match = re.search(r'\d{13}', element)
                    if account_number_match:
                        account_number = account_number_match.group()
                    # Split at '-' and remove '\n' to get the account type
                    parts = element.split('-')
                    if len(parts) > 1:
                        account_type = parts[1].replace('\n', '')    

            obj = {
                'name': '',
                'currency_code': '',
                'type': account_type,
                'iban': '',
                'account_number': account_number,
                'bank_name': 'Emirates Islamic Bank',
                'branch': '',
                'address': '',
                'account_balance': ''
            }

            new_lst = []
            for sublist in plain_text_data:
                for i, item in enumerate(sublist):
                    if item.startswith('STATEMENT DATE\nNARRATION\nCREDIT'):
                        new_lst.append(sublist[i + 1:])
                        break
            flattened_lst = [item for sublist in new_lst for item in sublist]
            split_lst = [item.split('\n')[:-1] for item in flattened_lst]

            # Create a new list with sublists by combining matching elements
            lst_1 = []
            temp_sublist = []

            for sublist in split_lst:
                if all(re.match(r'^[+-]?\d{1,3}(,?\d{3})*(\.\d{2})?$', elem) for elem in sublist):
                    temp_sublist += sublist
                else:
                    if temp_sublist:
                        lst_1.append(temp_sublist)
                    temp_sublist = sublist

            # Add the last sublist to the new list
            if temp_sublist:
                lst_1.append(temp_sublist)

            data_dicts = []
            for sublist in lst_1:
                data_dict = {
                    'date': sublist[0],
                    'description': ' '.join(sublist[1:-4]),
                    'credit': (float(sublist[-3].replace(',', '')) if re.match(r'^[+-]?\d{1,3}(,?\d{3})*(\.\d{2})?$', sublist[-3]) else 0.0),
                    'debit': -(float(sublist[-2].replace(',', '')) if re.match(r'^[+-]?\d{1,3}(,?\d{3})*(\.\d{2})?$', sublist[-2]) else 0.0),
                    'running_balance': (float(sublist[-1].replace(',', '')) if re.match(r'^[+-]?\d{1,3}(,?\d{3})*(\.\d{2})?$', sublist[-1]) else 0.0)
                }
                data_dicts.append(data_dict)

            # Create a DataFrame from the list of dictionaries
            df = pd.DataFrame(data_dicts)
            df['amount'] = df['credit'] + df['debit']
            df.drop(columns={'credit', 'debit'}, inplace=True)
            json_data = df.to_json(orient='records')
            return json_data
        except Exception as e:
            return str(e)


import Incom_expense
import BanksExtractor


def pdf_to_bytes(pdf_file_path):
    try:
        with open(pdf_file_path, "rb") as pdf_file:
            pdf_bytes = pdf_file.read()
        return pdf_bytes
    except FileNotFoundError:
        return None

pdf_file_path = 'abdalla.mharm@gmail.com.pdf'
pdf_bytes = pdf_to_bytes(pdf_file_path)

print('starting test')
res_json = BanksExtractor.BankExtractor().extract(pdf_bytes)
print(res_json)




import time
import ast
from scipy import stats
import difflib
import keyword
import warnings
import datetime
import json
import os
import warnings
from nltk.tokenize import word_tokenize
import numpy as np
import pandas as pd
import calendar
import re
import pkg_resources
pd.set_option('display.max_columns', None)

start = time.time()

from scipy.stats import pearsonr
from tqdm import tqdm
import tensorflow as tf
from nltk.tokenize import word_tokenize
from pickle import load
from sklearn.preprocessing import QuantileTransformer
from pandas.core.indexes.multi import MultiIndex
from collections import OrderedDict

from numpy import linalg
from pickle import dump

#for downloading BERT
from sentence_transformers import SentenceTransformer

#for finding most similar text vectors
from sklearn.metrics.pairwise import cosine_similarity

# Get the directory of the current script
script_dir = os.path.dirname(__file__)
# Construct the path to the CSV file
csv_file_path = os.path.join(script_dir, 'models')

labels = pd.read_csv(csv_file_path+'/lean/labels')['0']
model_path = pkg_resources.resource_filename('bankstatementextractor_final', 'models/best.pt')

class Incom_expense:   
    def __init__(self):
        pass

    def clean_text(self,text):

        text = re.sub(
            r'(https?(:\/\/)? ?)(www[\.]*)?', '',
            flags=re.IGNORECASE,
            string=re.sub(
                r'[^A-Za-z0-9&%\(\) .]+', '',
                re.sub(
                    r'[\*;:,\-\/\\\[\]_]+', ' ',
                    re.sub(
                        r'-?\d*\.?([\dA-Z]*)?\d+(?!TH|th|ND|nd|ST|st)', '',
                        string=re.sub(
                            r'(^|(?<=[^A-Za-z]))al ', '',
                            flags=re.IGNORECASE,
                            string=re.sub(
                                r'AE *\d[\dA-Z]*', '',
                                re.sub(
                                    r'DMCC', '',
                                    re.sub(
                                        r'LLC', '',
                                        re.sub(
                                            r'NFC[^A-Za-z]', '',
                                            re.sub(
                                                r'SMARTDXBGOV-', '',
                                                flags=re.IGNORECASE,
                                                string=re.sub(
                                                    r'^FAB -', '',
                                                    re.sub(
                                                        r'BNK REF\.', '',
                                                        re.sub(
                                                            r'TXNID\.', '',
                                                            re.sub(
                                                                r'DEBIT', '',
                                                                re.sub(
                                                                    r'V DEBIT', '',
                                                                    re.sub(
                                                                        r'Card Ending with', '',
                                                                        re.sub(
                                                                            r'Visa Purchase', '',
                                                                            re.sub(
                                                                                r'Visa Refund', '',
                                                                                re.sub(
                                                                                    r'Value Date', '',
                                                                                    re.sub(
                                                                                        r'PURCHASE TRXN', '',
                                                                                        re.sub(
                                                                                            r'INTERNET BANKING', '',
                                                                                            re.sub(
                                                                                                r'OUTWARD TT', '',
                                                                                                re.sub(
                                                                                                    r'CARD NO', '',
                                                                                                    re.sub(
                                                                                                        r'PUR ', '',
                                                                                                        re.sub(
                                                                                                            r'CARD TRANSACTION', '',
                                                                                                            re.sub(
                                                                                                                r'DR ATM TRANSACTION', '',
                                                                                                                re.sub(
                                                                                                                    r'OUTWARD UAE FUNDS TRANSFER - IPI', '',
                                                                                                                    re.sub(
                                                                                                                        r'ELECTRON DEBIT CARD TRANSACTION', '',
                                                                                                                        re.sub(
                                                                                                                            r'POS-PURCHASE;CARD NO', '',
                                                                                                                            re.sub(
                                                                                                                                r':AE', '',
                                                                                                                                re.sub(
                                                                                                                                    r'RAK:AE', '',
                                                                                                                                    re.sub(
                                                                                                                                        r'AJMAN', '',
                                                                                                                                        flags=re.IGNORECASE,
                                                                                                                                        string=re.sub(
                                                                                                                                            r'DUBAI', '',
                                                                                                                                            flags=re.IGNORECASE,
                                                                                                                                            string=re.sub(
                                                                                                                                                r'DUBAI:AE', '',
                                                                                                                                                re.sub(
                                                                                                                                                    r'DUBAI ARE', '',
                                                                                                                                                    re.sub(
                                                                                                                                                        r'DXB ARE', '',
                                                                                                                                                        re.sub(
                                                                                                                                                            r'DUBAI *(AE)|(UAE)|(ARE)|(DXB)', '',
                                                                                                                                                            re.sub(
                                                                                                                                                                r'AL AIN', '',
                                                                                                                                                                re.sub(
                                                                                                                                                                    r'AL AIN ARE', '',
                                                                                                                                                                    re.sub(
                                                                                                                                                                        r'AL AIN *(AE)|(UAE)|(ARE)', '',
                                                                                                                                                                        re.sub(
                                                                                                                                                                            r'AL AIN:AE', '',
                                                                                                                                                                            re.sub(
                                                                                                                                                                                r'Ras Al Khaima *(AE)|(UAE)|(ARE)|(RAK)', '',
                                                                                                                                                                                flags=re.IGNORECASE,
                                                                                                                                                                                string=re.sub(
                                                                                                                                                                                    r'FUJAIRAH', '',
                                                                                                                                                                                    re.sub(
                                                                                                                                                                                        r'FUJAIRAH *(AE)|(UAE)|(ARE)|(FUJ)', '',
                                                                                                                                                                                        flags=re.IGNORECASE,
                                                                                                                                                                                        string=re.sub(
                                                                                                                                                                                            r'SHARJAH', '',
                                                                                                                                                                                            flags=re.IGNORECASE,
                                                                                                                                                                                            string=re.sub(
                                                                                                                                                                                                r'SHARJAH ARE', '',
                                                                                                                                                                                                re.sub(
                                                                                                                                                                                                    r'SHARJAH:AE', '',
                                                                                                                                                                                                    re.sub(
                                                                                                                                                                                                        r'SHARJAH *(AE)|(UAE)|(ARE)|(SHJ)', '',
                                                                                                                                                                                                        re.sub(
                                                                                                                                                                                                            r'ABU *DHABI', '',
                                                                                                                                                                                                            flags=re.IGNORECASE,
                                                                                                                                                                                                            string=re.sub(
                                                                                                                                                                                                                r'ABU *DHABI ARE', '',
                                                                                                                                                                                                                re.sub(
                                                                                                                                                                                                                    r'ABU *DHABI:AE', '',
                                                                                                                                                                                                                    re.sub(
                                                                                                                                                                                                                        r'ABU *DHABI *(AE)|(UAE)|(ARE)|(AUH)', '',

                                                                                                                                                                                                                        re.sub(
                                                                                                                                                                                                                            r'BHD', '',
                                                                                                                                                                                                                            re.sub(
                                                                                                                                                                                                                                r'BHD ARE', '',
                                                                                                                                                                                                                                re.sub(
                                                                                                                                                                                                                                    r'BHD', '',
                                                                                                                                                                                                                                    re.sub(
                                                                                                                                                                                                                                        r'AED', '',
                                                                                                                                                                                                                                        str(text).strip(
                                                                                                                                                                                                                                        )
                                                                                                                                                                                                                                    )))))))))))))))))))))))))))))))))))))))))))))))))))))))
        ).strip()

        words = word_tokenize(text)
        result = ' '.join(words)
        return result


    def month_bound(self,x):
        return min(calendar.monthrange(x.year, x.month)[1] - x.day, x.day)

    def consecutive(self,numbers):
        #     numbers = pd.to_datetime(numbers)
        numbers = numbers.drop_duplicates().sort_values()
        max, count_ = 1, 1
        start_idx, end_idx = 0, 0
        for i in range(len(numbers) - 1):
            # if difference between number and his follower is 1,they are in sequence
            if numbers.iloc[i + 1] - numbers.iloc[i] == 1:
                #             print(i)
                count_ = count_ + 1
            else:
                if count_ > max:
                    max = count_
                    end_idx = i
                    start_idx = i + 1 - max
                # Reset counter
                count_ = 1
        return (
            (numbers.iloc[i + 2 - count_], numbers.iloc[i + 1], count_)
            if (max == 1 and count_ != 1)
            else None
            if (max == 1 and count_ == 1)
            else (numbers.iloc[start_idx], numbers.iloc[end_idx], max)
        )


     
    def get_category(self,x):
        global labels
        argmax = np.argmax(x)
        if argmax >= 0 and argmax < len(labels):
            return labels[argmax]
        else:
            return 'Unknown'


    def contains_salary(self,keyword_set):
        return 'salary' in keyword_set   


    def income_detection(self,res_json):
            salary=0
            average_income=0
            model_name = 'places'
            #models
            id = pd.Series(pd.to_numeric(os.listdir(csv_file_path+'/'+model_name), errors='coerce'),).dropna(

            ).sort_values(ascending=False).iloc[0]  
            #models
            places_categories = pd.read_csv(csv_file_path+'/categories.csv',)['0']
            # categories
            places_model = tf.keras.models.load_model(csv_file_path+'/'+model_name+'/'+str(id))

            model_name = 'words'
            id = pd.Series(pd.to_numeric(os.listdir(csv_file_path+'/'+model_name), errors='coerce'),).dropna(

            ).sort_values(ascending=False).iloc[0] 
            # .sort_values()
            words_categories = pd.read_csv(csv_file_path+'/categories.csv',)['0']
            # categories
            words_model = tf.keras.models.load_model(csv_file_path+'/'+model_name+'/'+str(id))

        # try:
            transactions = pd.DataFrame.from_records(ast.literal_eval(res_json))
            transactions_ = transactions[[
                'account_id','name','type', 'iban', 'account_number', 'currency_code', 'bank_name', 'branch', 'address']].drop_duplicates()
            transactions = transactions.rename(columns={
                'name': 'email',
                'account_number': 'account',
            }).drop(['type', 'iban','currency_code', 'bank_name'], axis=1)

            for col in ['description', 'amount', 'currency_code', 
                        'pending',
                        'timestamp', 'account', 'email', 'running_balance',]:
                if(not col in transactions.columns):
                    transactions[col] = pd.Series()
            transactions['timestamp']=transactions['timestamp'].apply(str)
            txs = transactions[(transactions['amount'] < 0)]
            credit = transactions[(transactions['amount'] > 0)]

            transactions_ = transactions_.rename(
                columns={
                    'name':'email'
                }
            )
            accounts = transactions_

            transactions_['account_id'] = transactions_['account_number']
            for col in ['account_id', 'name', 'currency_code', 'type', 'iban', 'account_number',
                        'credit', 'email', 'credit/next_payment_due_amount',
                        'credit/next_payment_due_date']:
                if(not col in transactions_.columns):
                    transactions_[col] = pd.Series()

            accounts['name']=accounts['name'].fillna('')
            accounts["credit"] = accounts["credit"].abs()
            accounts["credit/next_payment_due_amount"] = accounts[
                "credit/next_payment_due_amount"
            ].abs()


            txs['description']=txs['description'].str.strip('#')



            #TODO SmartDXB should be here too
            for prefix in '''GOOGLE \*
            GOOGLE\*
            Microsoft\*
            MICROSOFT\*
            PAYPAL \*
            Payup\*'''.split('\n'):
                txs.loc[#(txs['prefix2'].isna()) &
                        (
                    txs['description'].str.contains(prefix)), 'prefix2'] = prefix


            ## ;
            txs.loc[txs['prefix2'].isna(), 'prefix2'] = txs['description'].apply(lambda x: x.split(';')[0].strip() if (x.split(';')[0]).strip() in '''POS-PURCHASE
            \"POS-PURCHASE
            ONLINE BANKING TRANSFER
            SERVICE CHGS-ATM
            UAE SWITCH WDL
            Transfer between own accounts
            TRANSFER
            DR ATM TRANSACTION
            TELEGRAPHIC TRF
            CREDIT CARD PAYMENT
            DIRECTREMIT
            ATM WITHDRAWAL
            ONLINE BANKING PYMT-TEL
            ONLINE BANKING PYMT
            SALIK PAYMENT
            CLEARING CHEQUE
            INWARD REJECT
            OUTWARD REMITTANCE
            DEWA PAYMENT
            BANKNET TRANSFER
            DIRECT REMITTANCE
            SEWA PAYMENT
            SERVICE CHARGES
            UAE SWITCH ATM WITHDRAWAL
            TELECOM PAYMENT
            Cardless Withdrawal
            INTEREST COLLECTION
            INWARD CLEARING
            UAE SWITCH ENQUIRY CHRG
            Cheque Book Charges
            Online Acct to Acct transfer
            Corr.Bank.Charges
            Funds Transfer
            Charges
            PRINCIPAL DB. TO RETAIL
            INTEREST DB. TO RETAIL
            SUKUK FIN. INSTALLMENT PAYMENT
            INSTALLMENT PLAN EM
            OVERLIMIT FEE CHARGE
            CLEARING CHEQUES
            CHEQUE RETURNED
            LOAN DRAWDOWN
            STANDING ORDER
            WESTERN UNION MOBILE APPS
            INWARD CHQ RETURN
            PURCHASE
            CLEARING CHQ
            ONLINE NOQODI TOP UP
            CHARGE NOQODI TOP UP
            VALUE ADDED TAX
            CASH ADVANCE FEE
            LIV TRANSFER
            EIP INSTALLMENT
            INSTALLMENT PLAN PROCESSING FEE
            FORECLOSURE DEBIT TRANSACTION
            REPLACEMENT CARD FEE
            LOAN CHARGES
            LOAN SETTLEMENT
            LOAN PAYMENT
            LOAN ON CARD PROCESSING FEE
            LOAN ON CARD
            ITM CHEQUE ENCASHMENT
            INTEREST
            COUNTER CHEQUE
            ACCOUNT TRANSFER
            OUTWARD REJECT
            TRANSFER BETWEEN OWN ENBD ACCOUNTS'''.split('\n')
                                                                else np.nan)


            ## *
            #TODO: allow for whee GOOGLE* is not at the beginning
            txs.loc[txs['prefix2'].isna(), 'prefix2'] = txs[txs['prefix2'].isna()
                                                            ]['description'].apply(lambda x: x.split('*')[0].strip() if x.split('*')[0].strip() in '''REVOLUT
            MS
            SharePay
            SHAREPAY
            PayUp
            PAYUP
            MAMO PAY
            FBPAY
            JPAY
            Payex
            ADVANCED MPAY
            PAY
            PAYMOB
            PSP
            PTB
            GOOGLE'''.split('\n') else np.nan)


            ## :

            txs.loc[txs['prefix2'].isna(),'prefix2'] = txs[txs['prefix2'].isna()
                                                        ]['description'].apply(lambda x: x.split(':')[0].strip() if x.split(':')[0].strip() in '''Visa Purchase
            Value Added Tax - Output
            ATM Cash Withdrawal
            ATM Usage Fees
            Online Local Fund Transfer
            Transfer to other FAB account
            Credit Card Payment
            ATM-SWITCH-WITHDRAWAL-CHARGE
            Funds Transfer Charges
            ATM CASH WDL CHG AT UAE SWITCH MEMBER BK
            ATM Withdrawal - FAB ATM
            \"PAYMENT RECEIVED, THANK YOU\"
            AED TRF Within UAE Charges Incl VAT
            DDS Payment
            DU PAYMENT MBA
            Acct to Acct transfer
            Inward Clearing Cheque
            MB BILL DR DU
            MB DR
            Fall Below Fees
            Cardless Cash withdrawal
            Interest
            Breach of Min Balance
            Inward Clearing Return Cheque
            QuickRemit
            QuickRemit Charges
            ONLINE BANKING NOL TOP UP
            DIRECT DEBIT CENTRAL BANK UAEPGS
            Cheque Book Charges
            PRINCIPAL DB. TO RETAIL
            INTEREST DB. TO RETAIL
            SUKUK FIN. INSTALLMENT PAYMENT
            Inward Remittance
            Online International Money Transfer
            Corr.Bank.Charges
            WESTERN UNION MOBILE APPS
            Etisalat Bills
            Du Bills
            DEWA Payment
            MB BILL DR
            MBBILL DR
            IB BILL DR
            Relationship Fee
            ETISALAT PAYMENT MBA
            Inward Clearing Return Charges
            Principal Recovery
            Loan Recovery
            DEWA PAYMENT MBA
            SEWA PAYMENT MBA
            Card Auto Payment
            Standing Orders
            FEWA PAYMENT MBA
            DEWA PAYMENT WWW
            MRTA Insurance Charges
            Statement Charges
            Thank you.Card Payment
            Online Salik Payment
            Sharjah Electricity & Water Auth
            FTS CHARGES
            Instant Card Payment
            LOANRECOVERY-EMI
            #LOANRECOVERY-EMI
            DDS Rejection
            ETISALAT PAYMENT WWW
            Inward posting of cheque
            FEWA BILL PAYMENT
            AJMAN SEWERAGE PAYMENT MBA
            SALIK RECHARGE WWW
            Online Acct to Acct transfer
            Late Payment Fee Charges
            Abu Dhabi Distribution Company MBA
            Cheque Transaction
            FEWA PAYMENT WWW
            SEWA PAYMENT WWW
            Ma'an Donation
            Cash Advance
            Cheque Withdrawal
            Cheque
            Outward Clearing Return Cheque
            Cheque Return Charges
            Misc Charges
            Miscellaneous Charges
            NDRF Loan Recovery
            Profit Recovery
            Cover Account Transaction
            TELEGRAPHIC TRANSFER CHARGES
            Visa Card FEE
            Prepayment Charges
            IB BILL DR DU
            Card Payment by branch
            EMI Postponement Charges
            Al-Ain Payments'''.split('\n') else np.nan)





            for prefix in '''GOOGLE *
            GOOGLE*
            Microsoft*
            MICROSOFT*
            PAYPAL *
            Payup*
            INSTALLMENT PLAN EMI
            Local Transfer
            ELECTRON DEBIT CARD TRANSACTION
            OUTWARD UAE FUNDS
            Value Added Tax
            SWITCH ATM CASH WITHDRAWAL
            ATM CASH WITHDRAWAL
            Electron-Ref
            Social Transfer
            ATM WDL 
            PUR 
            POS-PURCHASE
            POS-
            POS Purchase
            POS 
            PURCHASE TRXN.
            Cash Withdrawal 
            SERVICE CHGS-ATM
            UAE SWITCH WDL
            Transfer to
            International Transfer to
            BILLED FINANCE CHARGES
            Cheque Paid
            Transfer 
            Debit Card Purchase
            \"PAYMENT RECEIVED, THANK YOU\"
            PAYMENT RECEIVED, THANK YOU
            VISA CARD PAYMENT
            MONTHLY ACCOUNT MAINTENANCE FEE
            SMS CHARGES
            SW WDL Chgs
            MB Fund Transfer Debit
            CARD TRANSACTION
            Charges Receivable Recovery
            General AE
            General S
            ATM-
            ATM withdrawal - Other Bank ATM
            LATE PAYMENT
            SWITCH FEE
            SMARTDXBGOV
            SmartDXBGov
            SMARTGOV
            SmartGov
            DIRECT DEBIT TXN - DUBAI EGOV
            VAT 
            \"Loan Transaction,
            Loan Transaction
            ATM BALANCE ENQUIRY CHARGE
            CHEQUE BOOK ISSUANCE
            CHARGE COLLECTION
            \"CHARGE COLLECTION
            ONLINE BANKING
            PAYROLL ACCOUNT MAINTENANCE FEES
            OUTWARD T/T
            STATEMENT CHARGES
            Profit earned
            Correspondent Bank Charges
            Corr.Bank.Charges
            FAB - FUND TRANSFER
            India Online Funds Transfer
            SMART DUBAI GOVERNMENT
            Smart Dubai Government
            OVERLIMIT FEE CHARGE
            OVERLIMIT CHARGE
            PAYMENT TO OWN CREDIT CARD
            CREDIT CARD PAYMNT
            Overlimit Fee
            UAEPGS Transaction
            SALIK RECHARGE MBA
            INTEREST 
            Minimum Balance Charge-
            CREDIT CARD PAYMENT
            Installment Recovery-
            LOAN RECOVERY
            Account Maintenance Fee
            IB FUNDS TRANSFER DEBIT
            UTILITY BILL PAYMENT
            CREDIT CARD RECOVERY
            MOB-CRDPAY
            Credit Card Payment
            CREDIT CARD AUTO PYMT
            OUTWARD TT
            TELEGRAPHIC TRF
            CASH MGMT. SERVICE CHARGES
            EXPRESS TRANSFER
            SwitchATM Withdrawal
            MOBN TRANSFER
            MOBN - INTERNAL TRANSFER
            MOBN - Card Payment
            Standing Order - Utility Bill Payment
            Standing Order
            Instant Transfer
            DEBIT CARD FEE
            TRANSFER International Payment
            TRANSFER AE
            Processing Fees
            PROCESSING FEES
            ATM DECLINE TRANSACTION CHARGE
            RETURNED CHEQUE
            Etisalat PAYMENT
            DU PAYMENT
            CLEARING CHEQUES
            INWARD DIRECT DEBIT
            Utility Bill Payment
            ATM Withdrawal
            PURCHASE 
            CORRESPONDENT BANK CHARGES
            OUTGOING CROSS BORDER TRANSFER
            Red Crescent
            Ma'an Donation
            MB BILL DR
            FUNDS TRANSFER BETWEEN OWN ACCOUNTS
            UAEDDS RTN CHARGES
            FUNDS TRANSFER
            BNKNT UTLTY PYMNT
            LOAN ON CALL EMI
            TFR REV IW IPI PAY REVERSAL
            ATM withdrawal
            ATM ADU Donation
            ATM 
            EPP Booking
            Outward Cheque Returned
            ES Fees
            PAYMENT TO OTHER RAKBANK CREDIT CARD
            Miscellaneous 
            TRANSFER REVERSAL 
            CHEQUE RETURN CHARGE
            DEWA PAYMENT
            ETISALAT PAYMENT
            LOANRECOVERY
            FINANCE CHARGES
            FINANCE CHARGE
            I/W CLEARING CHEQUE
            BILLED PROFIT
            Fund Transfer Charges
            Switch Charge-Txn
            TRANSFER TO 
            TRANSFER To
            Balance Enquiry-ATM
            Fees on Fund transfer 
            Fund transfer 
            DIRECT DEBIT RETURN FEE
            OVER CREDIT LIMIT FEES
            LEDGER FEES
            CBD ATM
            TRANSFER BETWEEN OWN
            ELECTRON CARD REPLACEMENT CHARGES
            Cheque Return Charge
            Utility Payment'''.split('\n'):
                txs.loc[(txs['prefix2'].isna())&(txs['description'].str.startswith(prefix)),'prefix2']=prefix




            for prefix in '''MBTRF
            TRF CORR CHARGE
            TRF :CORR CHARGE
            _IPI_
            ACCOUNT MAINTENANCE
            TEMPORARY HOLD
            TEMPORARYHOLD
            OUTGOING DOMESTIC FUND TRANSFER
            FUND TRANSFER
            Funds Transfer
            FTI-UAE-
            TOC-MAE-
            Payment - 
            TELEX TRANSFER
            CASH WITHDRAWAL
            CC Payment
            Sukuk Processing Fee
            ATM CASH TRANSFER
            Card Replacement
            Domestic Outward Remittance Charges
            PF Processing Fee
            IBTRF
            Domestic Outward Corr.Bank Charges
            GOOGLE \*
            GOOGLE\*'''.split('\n'):
                txs.loc[(txs['prefix2'].isna())&(txs['description'].str.contains(prefix)),'prefix2']=prefix


            txs['stem']=txs['description']
            txs.loc[~txs['prefix2'].isna(),'stem']=txs[~txs['prefix2'].isna()][['description','prefix2']].apply(
                lambda x:x['description'].replace(x['prefix2'].replace('\*','*'),'') ,axis=1)

            text_raw = txs['stem']
            text_BERT = text_raw.apply(self.clean_text)
            txs['clean'] = text_BERT


            purchases=txs[txs['prefix2'].isin([
                'Debit Card Purchase',
                'ELECTRON DEBIT CARD TRANSACTION',
                'POS-PURCHASE',
                'POS-',
                # 'POS-PURCHASE',
                'POS Purchase',
                'POS ',
                'PUR ',
                'PURCHASE TRXN.',
                'Visa Purchase',
                'GOOGLE*',
                'GOOGLE',
                'GOOGLE *',
                'SMARTDXBGOV',
                'SmartDXBGov',
                'SMARTGOV',
                'Smart Dubai Government',
            'SmartGov',
            'SMART DUBAI GOVERNMENT',
            'DIRECT DEBIT TXN - DUBAI EGOV',
            # 'OUTGOING DOMESTIC FUND TRANSFER',
            'FTI-UAE-',
                'TOC-MAE-',
                'Funds Transfer',
                'POS-',
                'Standing Orders',
                'STANDING ORDER',
                'IB FUNDS TRANSFER DEBIT',
                'Payment - ',
                'Microsoft*',
                'MICROSOFT*',
                'PAYPAL *',
                'Payup*',
                'REVOLUT',
                'Instant Transfer',
                'TRANSFER International Payment',
                'TRANSFER AE',
                'MS',
                'Standing Order',
                'PURCHASE',
                'SharePay',
            'SHAREPAY',
            'PayUp',
            'PAYUP',
            'MAMO PAY',
            'FBPAY',
            'JPAY',
            'Payex',
            'ADVANCED MPAY',
            'PAY',
            'PAYMOB',
                'PURCHASE ',

            'PSP',
            'PTB',
                'ACCOUNT TRANSFER',
                'TRANSFER TO ',
                'IBTRF',
                'TRANSFER To',
                'Fund transfer ',
                'GOOGLE \*',
                'GOOGLE\*',
                np.nan
            ])]


            purchases['month_year'] = purchases['timestamp'].apply(
                lambda x: str(x)[:7])

            more_cat = txs.drop(purchases.index)
            more_cat['category'] = more_cat['prefix2'].apply(lambda x: {

            }.get(x, 'Others'))

            purchases['merchant'] = purchases['clean'].dropna().str.lower(
            )
            purchases['words'] = purchases['merchant'].str.split(' ')

            data_dict = pd.DataFrame()
            data_dict['description'] = purchases['merchant'].unique()

            ## Lighter 
            bert_input = data_dict['description'].str.lower().tolist()
            model = SentenceTransformer('paraphrase-mpnet-base-v2')
            embeddings = model.encode(bert_input, show_progress_bar=True)
            embedding_BERT = np.array(embeddings)   

            X = pd.DataFrame(embedding_BERT, index=data_dict['description'])

            test_x = X
            y_pred = words_model.predict(test_x)

            y_pred_binary = (y_pred > 0.5).astype(int)
            y_pred_binary = pd.DataFrame(
                y_pred_binary, columns=places_categories, index=test_x.index).drop(['Others'], axis=1)


            pred = pd.DataFrame(y_pred, columns=words_categories, index=test_x.index).apply(lambda x: places_categories[np.argmax(x)] if np.max(x) > 0.5 else np.nan  # 'Others'  # np.nan
                                                                                            , axis=1)

            purchases['category'] = purchases['merchant'].apply(
                lambda x: pred.loc[x])

            test_x = X
            y_pred = places_model.predict(test_x)

            y_pred_binary = (y_pred > 0.5).astype(int)
            y_pred_binary = pd.DataFrame(
                y_pred_binary, columns=places_categories, index=test_x.index).drop(['Others'], axis=1)


            pred = pd.DataFrame(y_pred, columns=places_categories, index=test_x.index).apply(lambda x: places_categories[np.argmax(x)] if np.max(x) > 0.7 else np.nan  # 'Others'  # np.nan
                                                                                            , axis=1)

            purchases.loc[purchases['category'].isna(), 'category'] = purchases[purchases['category'].isna()]['merchant'].apply(
                lambda x: pred.loc[x])

            purchases['category'] = purchases['category'].fillna('Others')



            txs = credit.rename(columns={'amount_x': 'amount'}).copy()

            accounts_bytheemail = (
                accounts.groupby("email")
                .agg(
                    {
                        "credit": [
                            "sum",
                            #                   'count','mean','min','max'
                        ],
                        "credit/next_payment_due_amount": "sum",
                        "account_number": "nunique",
                        "type": "nunique",
                        "iban": "nunique",
                        "credit/next_payment_due_date": lambda x: (
                            pd.to_datetime(x.dropna(), errors="coerce")
                            .dropna()
                            .apply(self.month_bound)
                            .mean()
                        ),
                        "name": [
                            lambda x: 1 if x.str.lower()
                            .str.contains("liv |neo basic|no min balance")
                            .sum()>0 else 0,
                            lambda x: 1 if x.str.lower().str.contains("neo ").sum()>0 else 0,
                            lambda x: 1 if x.str.lower()
                            .str.contains("platinum|titanium|prime|gold|club|elite|infinite")
                            .sum()>0 else 0,
                            lambda x: 1 if x.str.lower().str.contains("special|classic").sum()>0 else 0,
                            lambda x: 1 if x.str.lower().str.contains("credit").sum()>0 else 0,
                            lambda x: 1 if x.str.lower().str.contains("cashback").sum()>0 else 0,
                            lambda x: 1 if x.str.lower().str.contains("standard").sum()>0 else 0,
                            lambda x: 1 if x.str.lower().str.contains("salary|salaried|pay roll").sum()>0 else 0,
                            lambda x: 1 if x.str.lower().str.contains("current").sum()>0 else 0,
                            lambda x: 1 if x.str.lower().str.contains("investment|trader|trading").sum()>0 else 0,
                            lambda x: 1 if x.str.lower().str.contains("savings|save |saver|s@ver").sum()>0 else 0,
                        ],
                    }
                )
                .reset_index()#.astype('float')
            )
            try:

                accounts_bytheemail["credit_ratio"] = (
                accounts_bytheemail[("credit/next_payment_due_amount", "sum")]
                / accounts_bytheemail[("credit", "sum")]
                ).abs()

            except: 
                accounts_bytheemail["credit_ratio"] = ''

            accounts_bytheemail[("credit_ratio")] = accounts_bytheemail[("credit_ratio")].replace(
                np.inf, 0
            )
            accounts_bytheemail[("credit_ratio")] = accounts_bytheemail[("credit_ratio")].fillna(0)




            txs['timestamp'] = txs['timestamp'].astype('str')

            recalculate = True
            txs["abs_amount"] = txs["amount"].abs()

            all_dup = txs[
                txs.duplicated(
                    subset=["abs_amount", "email", "timestamp", "description"], keep=False
                )
            ]
            txs.drop("abs_amount", axis=1, inplace=True)
            zero_sum = all_dup.drop(
                txs[
                    txs.duplicated(
                        subset=["amount", "email", "timestamp", "description"], keep=False
                    )
                ].index
            )
            zero_sum  # .head(20)

            txs = txs.drop(zero_sum.index)
            txs
            list_of_months = [
                "january",
                "february",
                "march",
                "april",
                "may",
                "june",
                "july",
                "august",
                "september",
                "october",
                "november",
                "december",
            ]
            list_of_months.extend([m[:3] for m in list_of_months])

            list_of_local = [
                "AE",
            ]
            list_of_foreign = ["UK", "US"]  # clashes with us(e.g HOMES R US)
            recent_years = ["2021", "2022", "2023"]

            list_of_online = [r"\.com", "www", "http"]


            keywords = pd.DataFrame(index=txs.index)
            description = txs["description"].str.replace(";", "").str.lower()

            tags = {
                "wage": description.str.contains("wage"),
                "compensation": description.str.contains("compensat"),
                "salary": description.str.contains("salary")
                | (txs["description"].str.contains("WPS|HR") & description.str.contains("sal"))
                | txs["description"].str.contains(
                    r"(?:^|[^a-zA-Z])({})(?=[^a-zA-Z]|$)".format("|".join(["SAL"]))
                ),

                "HR": txs["description"].str.contains(
                    r"(?:^|[^a-zA-Z])({})(?=[^a-zA-Z]|$)".format("|".join(["HR", "HRD"]))
                ),  # ).apply(len)>0,
                "allowance": description.str.contains("allowance")
                | (txs["description"].str.contains("HR") & description.str.contains("allow")),
                "advance": description.str.contains("advance([^a-zA-Z]|$)")
                | (txs["description"].str.contains("HR") & description.str.contains("adv")),
            "bill": description.str.contains(
                    r"(?:^|[^a-zA-Z])({})(?=[^a-zA-Z]|$)".format("|".join(["bill"]))
                ),  # ).apply(len)>0,
                "goods_services": description.str.contains("goods")
                & description.str.contains("bought|sold")
                | (description.replace(" ", "").str.contains("supplier payment")),
                "professional": description.str.contains("professional"),
                "bonus": description.str.contains("bonus")
                & description.str.contains(
                    "performance|hr|payment"
                ),  # TODO: specifically Perfomance buns only
                "overtime": description.str.contains("overtime"),
                "govt": description.str.contains("(gov(?=[^a-zA-Z]|$))|government|govt|ministry"),
                "incentive": description.str.contains("incentive"),
                "deposit": description.str.contains("deposit"),
                "IPI": txs["description"].str.contains("IPI transaction"),
                "cheque": description.str.contains("cheque") | description.str.contains("chq"),
                "cash": description.str.contains("cash") & (~description.str.contains("cashback")),
                "purchase": description.str.contains("purchase")
                | (
                    (txs["description"].str.slice(0, 4) == "PUR ")
                    & (description.str.contains("amazon"))
                ),
                "return": description.str.contains("return")
                | (
                    txs["description"].str.contains("DDS")
                    & (txs["description"].str.contains("DDR|NSF"))
                ),
                "refund": description.str.contains("refund")
                | (
                    (txs["description"].str.slice(0, 4) == "REF ")
                    & (description.str.contains("amazon"))
                ),
                "reversal": description.str.contains("reverse|reversal"),
                "reimburse": description.str.contains("reimburse|reimurse"),
                "claim": description.str.contains("claim"),
                "adjust": description.str.contains("adjustment|credit adj"),
                "card_tx": description.str.contains("card")
                & description.str.contains("transaction|txn|settlement"),
                "electron": description.str.contains(
                    r"(?:^|[^a-zA-Z])({})(?=[^a-zA-Z]|$)".format("|".join(["electron"])))
                & description.str.contains("ref|debit|card|visa"),
                "transfer": description.str.contains("transfer")
                | txs["description"].str.contains("/REF/"),
                "remittance": description.str.contains("remittance"),
                "inward": description.str.contains("inward")
                & (~description.str.contains("remittance")),
                # TODO: add not remittance
                "outward": description.str.contains("outward"),
                "commission": description.str.contains("commission"),
                "payment": description.str.contains("paymen|pymt"),
                "installment": description.str.contains("installment"),
                "rent": description.str.contains("rent") | txs["description"].str.contains("RNT"),
                "transport": description.str.contains("transport"),
                #             |txs['description'].str.contains('RNT'),
                "closure": description.str.contains("closure"),
                "fee": description.str.contains("fee"),
                "interest": description.str.contains("interest"),
                "profit": description.str.contains("profit"),
                "investment": description.str.contains("investm|invest "),
                "financial": description.str.contains("financ|cashback"),
                "saving": description.str.contains("saving|rainy day"),
                "pension": description.str.contains("pension"),
                "verification": description.str.contains("verif"),
                "stand_order": description.str.contains("standing order"),
                "temporary": description.str.contains("temporary")
                | (description.str.contains("temp") & description.str.contains("credit|adj")),
                "insufficient": description.str.contains("insufficient funds"),
                "maid": description.str.contains("maid |maids"),
                "company": description.str.contains(
                    "fze|fzc|llc|fzllc|fz.llc|llc.fz|l l c|l\.l\.c|dmcc|gmbh|ltd|limited|inc([^a-zA-Z]|$)"
                ),
                "benefit": description.str.contains("end.*service.*benefit")
                | description.str.contains("benefits"),
                "family": description.str.contains("family")
                | txs["description"].str.contains(
                    r"(?:^|[^a-zA-Z])({})(?=[^a-zA-Z]|$)".format("|".join(["FAM"]))
                ),
                "loan": description.str.contains("loan|lend|borrow"),
                "credit": description.str.contains("credit"),
                "drawdown": description.str.contains("drawdown"),
                "support": description.str.contains("support"),
                "educ": description.str.contains("education")
                | description.str.contains(
                    r"(?:^|[^a-zA-Z])({})(?=[^a-zA-Z]|$)".format("|".join(["educ", "edu"]))
                ),
                #     'B/O':txs['description'].str.slice(0,3)=='B/O',
                "bonds": description.str.contains("bonds.*redemption"),
                "withdraw": description.str.contains("withdraw")
                | (txs["description"].str.contains("SWITCH WDL")),
                "charge": description.str.contains("charge|corr chg")
                | (txs["description"].str.contains("SERVICE CHG")),
                "brought_over": txs["description"].str.contains(
                    "(^|[^a-zA-Z0-9])B/O([^a-zA-Z0-9])"
                ),
                "FTS": txs["description"].str.contains("FTS"),
                "ATM": txs["description"].str.contains("ATM"),
                "CDM": txs["description"].str.contains("CDM"),
                "telegraphic": txs["description"].str.contains("T/T"),
                "IPI": txs["description"].str.contains(
                    "IPI"
                ),  # TODO: watch out, there's something called IPI payment
                "DDS": txs["description"].str.contains(
                    r"(?:^|[^a-zA-Z])({})(?=[^a-zA-Z]|$)".format("|".join(["DDS"]))
                ),
                #     .str.contains('DDS'),
                #     &(~(txs['description'].str.contains('DDR|NSF'))),
                "invoice": txs["description"].str.contains(
                    r"(?:^|[^a-zA-Z])({})(?=[^a-zA-Z\.][^$a-zA-Z])".format("|".join(["INV"]))
                )
                | description.str.contains("invoice"),
                "POS": txs["description"].str.contains(
                    r"(?:^|[^a-zA-Z])({})(?=[^a-zA-Z]|$)".format("|".join(["POS"]))
                ),  # ).apply(len)>0,
                #     str.contains('POS'),
                "VAT": txs["description"].str.contains(
                    r"(?:^|[^a-zA-Z])({})(?=[^a-zA-Z]|$)".format("|".join(["VAT"]))
                )
                | description.str.contains("value added tax"),
                "UAEFTS": txs["description"].str.contains("UAEFTS"),
                "DIBFTS": txs["description"].str.contains("DIBFTS"),
                #     'DISB':txs['description'].str.contains('DISB')|description.str.contains('disbursement'),
                "disburse":
                #      txs['description'].str.contains('DISB')|
                description.str.contains("disburs"),
                # TODO: watch this
                "mktplace_payer": description.str.contains(
                    "noon e commerce|noonfood|zomato online sales|facebook payout"
                ),
                "gig_payer": description.str.contains("airbnb|booking\.com|souq\.com"),
                "GPSSA": txs["description"].str.contains("GPSSA"),
                "wallet_cash_out": txs["description"].str.contains("MOBILE WALLET CASH OUT"),
                "self": description.str.contains("myself|own acc|my acc")
                | (txs["description"].str.contains("OAT")),
                "local": txs["description"].str.contains(
                    r"(?:^|[^a-zA-Z])({})(?=[^a-zA-Z]|$)".format("|".join(list_of_local))
                ),
                # ).apply(len)>0,
                "foreign": txs["description"].str.contains(
                    r"(?:^|[^a-zA-Z])({})(?=[^a-zA-Z]|$)".format("|".join(list_of_foreign))
                ),
                # ).apply(len)>0,
                "online": description.str.contains(r"({})".format("|".join(list_of_online))),
                # ).apply(len)>0,
                "month": description.str.contains(
                    r"(?:^|[^a-zA-Z0-9])({})(?=[^a-zA-Z]|$)".format("|".join(list_of_months))
                )
                # ).apply(len)>0)
                | description.str.contains(
                    r"((?:^|[^0-9-\/])([0-1][0-9])(-|\/)?(2023|2022))|((2023|2022)(-|\/)?([0-1][0-9])(?=[^0-9-\/]|$))".format(
                        "|".join(recent_years)
                    )
                )
                # ).apply(len)>0)
            } if recalculate else {}
            for tag in tags:
                keywords.loc[tags[tag], tag] = tag

            if(recalculate):
                keywords.loc[(~keywords["purchase"].isna()) & (
                    txs["amount"] > 0), "refund"] = "refund"
                keywords.loc[
                    (~keywords["DIBFTS"].isna()) & (
                        txs["description"].str.slice(0, 5) == "Bonus"),
                    "bonus",
                ] = "bonus"
                keywords.loc[
                    (~keywords["DIBFTS"].isna()) & (
                        txs["description"].str.slice(0, 9) == "Allowance"),
                    "allowance",
                ] = "allowance"

                keywords.loc[txs["description"].str.slice(
                    0, 4) == "REF ", "refund"] = "refund"
                keywords.loc[txs["description"].str.slice(
                    0, 4) == "PUR ", "purchase"] = "purchase"
                keywords.loc[(~keywords["saving"].isna()) &
                            keywords["transfer"], "self"] = "self"

                keywords.to_csv('keywords.csv', index=False)
            else:

                keywords = pd.read_csv('/keywords.csv')

            keywords

            txs["keywords"] = keywords.agg(
                lambda x: ",".join(x.dropna()), axis=1).fillna('')


            # for tag in tqdm(tags):
            #     txs[txs["keywords"].str.contains(tag)][txs["amount"] > 0].to_csv(
            #         "keywords/" + tag + ".csv"
            #     )

            # txs[(txs["keywords"] == "") | txs["keywords"].isna()][txs["amount"] > 0].to_csv(
            #     "no_keywords.csv"
            # )


            txs["month"] = txs["timestamp"].str.slice(5, 7).astype("int")
            txs["day"] = txs["timestamp"].str.slice(8, 10).astype("int")
            txs["date"] = txs["timestamp"].str.slice(0, 10)  # .astype('int')

            # txs[txs['tags']>1]
            txs["keywords"] = txs["keywords"].fillna('')
            txs[txs["keywords"].str.contains("foreign")][txs["amount"] > 0]
            txs[txs["keywords"].str.contains("local")][
                ~txs["keywords"].str.contains(
                    "|".join(
                        [
                            "transfer",
                            "purchase",
                            "withdraw",
                            "salary",
                            "IPI",
                            "fee",
                            "ATM",
                            "self",
                            "refund",
                        ]
                    )
                )
            ][txs["amount"] > 0]

            txs[txs["keywords"].str.contains("online")]["description"].values
            txs[(txs["email"] == "ralzaabi555@gmail.com") & (txs["amount"] > 0)]


            spend = (
                txs[
                    txs["amount"]
                    < 0
                    #           &txs['purchase']
                ]
                .groupby(["email", "month"])["amount"]
                .sum()
            )
            # spend
            credit = txs[
                (txs["amount"] > 0)
                #              & (~txs['refund']) & (~txs['payment']) & (~txs['interest'])
            ]  
            throw_away = credit[
                credit["keywords"].str.contains("purchase")
                | (
                    credit["keywords"].str.contains("return")
                    & credit["keywords"].str.contains("outward|cheque|DDS")
                )
                | credit["keywords"].str.contains("refund")
                | credit["keywords"].str.contains("reversal")
                | credit["keywords"].str.contains("adjust")
                | credit["keywords"].str.contains("card_tx")
                | credit["keywords"].str.contains("electron")
                | (
                    credit["keywords"].str.contains("outward")
                    & credit["keywords"].str.contains("remittance|cheque")
                )
                | (credit["keywords"].str.contains("rent") & credit["keywords"].str.contains("IPI"))
                | (credit["keywords"].str.contains("closure"))
                | (
                    credit["keywords"].str.contains("financial")
                    & (
                        credit["keywords"].str.contains("disburse|charge")
                        | credit["description"].str.lower().str.contains("cashback")
                    )
                )
                | (credit["keywords"].str.contains("verification"))
                | (
                    credit["keywords"].str.contains("temporary")
                    & credit["keywords"].str.contains("credit")
                )
                | (credit["keywords"].str.contains("insufficient"))
                | (
                    credit["keywords"].str.contains("loan")
                    & credit["keywords"].str.contains("IPI|deposit")
                )
                | (credit["keywords"].str.contains("drawdown"))
                | (credit["keywords"].str.contains("withdraw"))
                | (credit["keywords"].str.contains("charge"))
                | (
                    credit["keywords"].str.contains("VAT")
                    #       &credit['keywords'].str.contains('charge|reversal')
                )
                | (
                    credit["keywords"].str.contains("disburse")
                    & credit["keywords"].str.contains("loan|financial")
                )
                | (credit["keywords"].str.contains("self"))
                | (
                    credit["keywords"].str.contains("online")
                    & credit["keywords"].str.contains("card_tx|refund|return|reversal|electron")
                )
            ]

            throw_away.to_csv("throw_away.csv", index=False)
            credit = credit.drop(throw_away.index, axis=0)

            credit["index"] = credit.index


            credit = credit.merge(spend.reset_index(), on=["email", "month"], how="left")
            credit.index = credit["index"]

            credit["perc"] = (credit["amount_x"] / credit["amount_y"].abs() * 100).round(1)



            emails = credit["email"].unique()
            len(emails)

            credit['date'] = pd.to_datetime(credit["date"])

            credit["index"] = credit.index
            samples = []
            groups = []
            counter = 0
            for email in (emails[:] if recalculate else []):
                #     ['zuraimaan@gmail.com']:#
                counter += 1

                sample = credit[credit["email"] == email]
                match = pd.DataFrame()
                for i in sample.index:
                    if i in sample.index:
                        sample["matcher"] = sample.apply(
                            lambda x: difflib.SequenceMatcher(
                                None,
                                re.sub(r"[0-9]+", "", sample.loc[i]
                                    ["description"].lower().replace(";", "")),
                                re.sub(r"[0-9]+", "",
                                    x["description"].lower().replace(";", "")),
                            ),
                            axis=1,
                        )
                        sample["score"] = sample["matcher"].apply(
                            difflib.SequenceMatcher.ratio)

                        matches = pd.concat(
                            [
                                (sample.head(1)),
                                sample.drop(i)[
                                    sample.drop(i).apply(
                                        lambda x: (x["score"] > 0.5  # 0.6#TODO: 0.5
                                                )  # and (abs((sample.loc[i]['day'] +(30 if sample.loc[i]['day']<15 else 0))
                                        #                             -(x['day'] +(30 if x['day']<15 else 0)))%30)<5
                                        ,
                                        axis=1,
                                    )
                                ],
                            ]
                        )          

                        sample = sample.drop(
                            matches.index
                            #                                ,inplace=True
                        )
                        if len(matches) == 1:
                            tx = matches.iloc[0]
                            tx['match_common'] = tx['description']
                            tx["group"] = i
                            tx["month_bound"] = (min(
                                calendar.monthrange(tx["date"].year, tx["date"].month)[
                                    1] - tx["date"].day, tx["date"].day
                            )
                            )
                            samples.append(
                                tx
                                #matches

                            )
                            continue

                        consistency = (
                            pd.to_datetime(matches["date"])
                            .sort_values()
                            .diff()
                            .apply(lambda x: x.days)
                        )
                        #             amt_consistency = (matches['amount_x']).sort_values().diff()#.apply(lambda x:x.days)
                        if len(matches["month"].unique()) > 1 or consistency.sum() >= 25:

                            matches["match_common"] = matches.apply(
                                lambda x: matches.loc[i]["description"][
                                    sorted(
                                        x["matcher"].get_matching_blocks(),
                                        key=lambda y: y[2],
                                        reverse=True,
                                    )[0][0]: sorted(
                                        x["matcher"].get_matching_blocks(),
                                        key=lambda y: y[2],
                                        reverse=True,
                                    )[
                                        0
                                    ][
                                        0
                                    ]
                                    + sorted(
                                        x["matcher"].get_matching_blocks(),
                                        key=lambda y: y[2],
                                        reverse=True,
                                    )[0][2]
                                ],
                                axis=1,
                            )
                            matches["match_diff"] = matches.apply(
                                lambda x: x["description"][
                                    sorted(
                                        x["matcher"].get_matching_blocks(),
                                        key=lambda y: y[2],
                                        reverse=True,
                                    )[0][1]
                                    + sorted(
                                        x["matcher"].get_matching_blocks(),
                                        key=lambda y: y[2],
                                        reverse=True,
                                    )[0][2]:
                                ],
                                axis=1,
                            )

                            matches["group"] = i
                            matches["month_bound"] = (matches["date"]).apply(
                                lambda x: min(
                                    calendar.monthrange(x.year, x.month)[1] - x.day, x.day
                                )
                            )
                            matches["consistency"] = consistency
                            groups.append(matches)


            print(groups)
            if len(groups) > 0 and groups:
                income = pd.concat([pd.concat(groups),
                                    # pd.concat(samples, axis=1).T
                                    ])
                income['iban'] = ''
                income['account_number'] = ''
                income['name'] = ''




                income["match_diff"].fillna("", inplace=True)
                income["keywords"].fillna("", inplace=True)
                # income['acc']=income['name'].str.join(income['bank'])
                income["bankid"] = income["iban"].str.slice(4, 7)

                # TODO: number of days to end of specific month
                income["months_ago"] = (income["date"]).apply(
                    lambda x: (
                        (datetime.datetime.today().year - x.year) * 12
                        + datetime.datetime.today().month
                        - x.month
                    )
                )
                income["months_ago"].value_counts()
                # income[['date','months_ago']]


                global_percentiles = [0.01, 50.0, 100.0, 220.0, 480.0, 700.0, 1134.0, 2000.0, 3500.0, 6950.0]

                # with open("Anjali_BS_Extract/bankstatementextractor_final/global_percentiles.json", "r") as f:
                #     # Write the dictionary to the file in JSON format
                #     global_percentiles = json.load(f)

                income['rank'] = (pd.cut((income["amount_x"]), bins=[0]+global_percentiles+[1e23],
                                        labels=range(len([0]+global_percentiles)),#duplicates='raise'
                                        ).astype('int')/len(global_percentiles)
                                )

                income_streams = (
                    income.groupby(["email", "group"])
                    .agg(
                        {
                            "amount_x": [
                                'sum',
                                "mean",
                                "count",
                                "min",
                                "max",
                                "std",
                                "median",
                                #         lambda x: x.quantile(0.25),lambda x: x.quantile(0.75)
                            ],
                            "rank": [
                                #         'mean',
                                "median",
                                #         lambda x: x.quantile(0.25),lambda x: x.quantile(0.75),
                                #         'min','max'
                            ],
                            #     'amount_x':,
                            "day": [
                                "mean",
                                #            pd.Series.mode
                            ],
                            "account_number": "first",
                            "month_bound": [
                                "mean",
                                #                    pd.Series.mode
                            ],
                            #     'date':consecutive,
                            "month": "nunique",
                            "months_ago": [
                                "min",
                                "max",
                                self.consecutive
                                #             '|'.join,
                            ],
                            "consistency": [
                                'sum',
                                #         'mean','std',
                                "median",
                                "mean",
                                "std",
                            ],

                            "match_common": "first",  # ' \ '.join,
                            "match_diff": " \ ".join,
                            "keywords": lambda x: set(",".join(x).split(",")),
                            "name": set,
                            "perc": ["mean", "std", "median"],
                            "bankid": set,
                        }
                    )
                    .merge(accounts_bytheemail, on="email", suffixes=["", "_user"])
                )


                income_streams['amount'] = income_streams.apply(lambda x: x[('amount_x', 'sum')]/(
                    ((x[('consistency', 'sum')])/(365.25/12)+1)) if x[('consistency', 'mean')] > 30 else x[('amount_x', 'sum')]/(
                    x[('months_ago', 'max')]-x[('months_ago', 'min')]+1), axis=1)

                # income_streams[('months_ago','max')]-income_streams[('months_ago','min')]+1)
                income_streams.drop([('amount_x', 'sum')], axis=1, inplace=True)
                income_streams.drop([('consistency', 'sum')], axis=1, inplace=True)
                # income_streams

                income_streams[("credit", "sum")].fillna(0, inplace=True)
                income_streams[("credit_ratio", "")].fillna(0, inplace=True)
                income_streams[("amount_x", "dev")] = (
                    income_streams[("amount_x", "std")] /
                    income_streams[("amount_x", "median")] * 100
                )
                income_streams[("consistency", "dev")] = (
                    income_streams[("consistency", "std")]
                    / income_streams[("consistency", "mean")]
                    * 100
                )
                income_streams["stability"] = (
                    income_streams[("consistency", "dev")] +
                    income_streams[("amount_x", "dev")]
                )
                # income_streams_filtered = income_streams[
                #     #     :
                #     (
                #         ~income_streams[("keywords", "<lambda>")].apply(
                #             lambda x:
                #             ("refund" in x)
                #         )
                #     )
                #     &
                #     (
                #         (income_streams["stability"] <= 103)
                #         | (
                #             (
                #                 (income_streams[("consistency", "dev")] < 20)
                #                 & (
                #                     (income_streams[("month_bound", "mean")] <= 6)
                #                 )

                #             )
                #             | (
                #                 (income_streams[("amount_x", "dev")] <= 25)
                #                 | (
                #                     (income_streams[("amount_x", "dev")] <= 47)
                #                     & (income_streams[("month_bound", "mean")] <= 6)
                #                 )
                #             )
                #         )
                #         | income_streams[("keywords", "<lambda>")].apply(
                #             lambda x:

                #             ("month" in x)
                #         )
                #     )

                # ]  

                income_streams.columns = [
                    "_".join(col).strip() if type(col) is tuple else col
                    for col in income_streams.columns.values
                ]

                columns_to_convert = ['credit_ratio_', 'credit_sum']

                income_streams[columns_to_convert] = income_streams[columns_to_convert].apply(pd.to_numeric, errors='coerce')

                # income_with_auth = income_streams.rename(columns={'email': 'email_'})
                keywds = pd.DataFrame(index=income_streams.index, columns=keywords.columns)
                for key in keywords.columns:
                    keywds[key] = income_streams[("keywords_<lambda>")].apply(
                        lambda x: key in x)


                desc_tags = pd.DataFrame(index=income_streams.index, columns=keywords.columns)

                desc_tags[keywds] = 1
                desc_tags[~keywds] = 0

                # category = "Business"
                desc = desc_tags

                income_train = income_streams.drop(
                    [
                        'email_',
                        'match_common_first',
                        'match_diff_join',
                        'keywords_<lambda>',
                        'name_set',
                        'months_ago_consecutive',
                        'bankid_set',
                        'credit/next_payment_due_date_<lambda>',
                        'account_number_first'
                    ],

                    axis=1,
                )
                

                qt = load(open('qt.pkl', 'rb'))
                income_train = pd.DataFrame(
                    qt.fit_transform(income_train),
                    columns=income_train.columns, index=income_train.index)

                data_x_train = pd.concat(
                    [
                        income_train  
                        ,
                        desc.astype('int'),
                        # nfc_cat.astype('int'),
                    ],
                    ignore_index=False,
                    axis=1,
                )



                fill_with = data_x_train.median(axis=0)

                for c in data_x_train.columns:
                    data_x_train[c].fillna(fill_with[c], inplace=True)


                fill_with.index = [
                    "_".join(col).strip() if type(col) is tuple else col
                    for col in fill_with.index.values
                ]


                for c in data_x_train.columns:
                    data_x_train[c].fillna(fill_with[c], inplace=True)

                model = tf.keras.models.load_model(
                    csv_file_path+'/lean/0.9384675110399379')
                # labels = pd.read_csv('models/lean/labels')['0']


                income_streams['category'] = pd.DataFrame(model.predict(data_x_train)).apply(self.get_category, axis=1)
                #income grouped
                income_streams['type'] = income_streams['category'].apply(
                    lambda x: 'Primary' if x == 'Salary' or x == 'Pension' else 'Secondary')
                income_streams['contains_salary'] = income_streams['keywords_<lambda>'].apply(lambda x: self.contains_salary(x))

                # Check if 'salary' is found in any of the sets
                if income_streams['contains_salary'].any():
                    salary = income_streams[income_streams['contains_salary']]['amount_'].mean()
                    average_income = income_streams.amount_.mean()


            spend_categories = {
                'Food & Drink': 'Variable',
                'Shopping': 'Variable',
                'Groceries': 'Fixed',
                'Others': 'Variable',
                'Transportation': 'Fixed',
                'Subscriptions': 'Fixed',
                'Rent & Utilities': 'Fixed',
                'Medical': 'Variable',
                'Online shopping': 'Variable',
                'Travel':  'Variable',
                'Transfer': 'Variable',
                'Entertainment': 'Variable',
                'Services': 'Variable',
                'Investment': 'Variable',
                'Leisure': 'Variable',
                'Bank fees': 'Variable',
                'Loans': 'Variable',
                'Income': 'Variable',

            }

            #expenses grouped
            purchases['spend_type'] = purchases['category'].apply(
                lambda x: spend_categories[x])

            grouped = purchases.groupby('category')['amount'].sum().reset_index()
            grouped['amount'] = grouped['amount'] * -1
            grouped['percentage'] = grouped['amount'] / -purchases['amount'].sum() * 100

            dict1 = grouped.to_dict(orient='records')

            combined_json = json.dumps({'expenses': dict1, 'salary': salary, 'average_income': average_income}, indent=2)
            print( f"first check:{combined_json}")
            return combined_json

        # except:
        #     #transactions_test2=type(transactions_test2)
        #     print( f"second check:{transactions_test.columns.to_list()}")
           

        



            



import base64
import cv2
import io
import numpy as np
import re
from datetime import datetime
from PIL import Image
from pdf2image import convert_from_bytes

from bankstatementextractor_final.banks_utils import *
from bankstatementextractor_final.banks import Banks
import bankstatementextractor_final.constants as const
from bankstatementextractor_final.Incom_expense import Incom_expense

# from banks_utils import *
# from banks import Banks
# import constants as const
# from Incom_expense import Incom_expense

import os
import subprocess
import torch 
os.sys.path
from io import BytesIO
import pkg_resources


class BankExtractor:

    def __init__(self):
        """
        This is the initialization function of a class that imports a spoof model and loads an BS
        extractor.
        """
        # self.bank_labels = ["adib_1","cbd_1","liv_1"]
    
    def __check_pdf_metadata(self, pdf_bytes):
        try:
            # Run the pdfinfo command with input as bytes
            # print('before run')
            pdfinfo_process = subprocess.run(["pdfinfo", "-"], input=pdf_bytes, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            if pdfinfo_process.returncode != 0:
                return False, "pdfinfo command failed."

            pdfinfo_output = pdfinfo_process.stdout
            # print(pdfinfo_output)
        except FileNotFoundError:
            return False, "pdfinfo command not found. Make sure it's installed and in your PATH."
        # Extract the creator and producer strings, created_at, and modified_at from the pdfinfo output
        creator_string = None
        producer_string = None
        created_at = None
        modified_at = None

        for line in pdfinfo_output.splitlines():
            line_str = line.decode('utf-8')  # Convert bytes to string
            if line_str.startswith("Creator:"):
                creator_string = line_str[len("Creator:"):].strip()
            elif line_str.startswith("Producer:"):
                producer_string = line_str[len("Producer:"):].strip()
            elif line_str.startswith("CreationDate:"):
                created_at = line_str[len("CreationDate:"):].strip()
            elif line_str.startswith("ModDate:"):
                modified_at = line_str[len("ModDate:"):].strip()

        crt_list_good= const.creator_lst_good
        crt_list_bad=const.creator_lst_bad
        prod_list_good=const.producer_lst_good
        prod_list_bad=const.producer_lst_bad
        # Check conditions based on the presence of strings in the lists and created/modified timestamps
        if creator_string in crt_list_good and producer_string in prod_list_good:
            return True, None
        elif creator_string in crt_list_good and producer_string in prod_list_bad:
            return False, "Upload a non-edited PDF"
        elif creator_string in crt_list_bad and producer_string in prod_list_good:
            return False, "Upload a non-edited PDF"
        elif creator_string in crt_list_bad and producer_string in prod_list_bad:
            return False, "Upload a non-edited PDF"
        elif creator_string not in crt_list_good + crt_list_bad and producer_string in prod_list_good:
            if created_at == modified_at:
                return True, None
            else:
                return False, "Upload a non-edited PDF"
        elif creator_string not in crt_list_good + crt_list_bad and producer_string in prod_list_bad:
            return False, "Upload a non-edited PDF"
        elif creator_string in crt_list_good and producer_string not in prod_list_good + prod_list_bad:
            if created_at == modified_at:
                return True, None
            else:
                return False, "Upload a non-edited PDF"
        elif creator_string in crt_list_bad and producer_string not in prod_list_good + prod_list_bad:
            return False, "Upload a non-edited PDF"
        else:
            if created_at == modified_at:
                return True, None
            else:
                return False, "Upload a non-edited PDF"
        
        # return result, message, metadata

    # Function to convert PDF to images using pdf2image library
    def convert_pdf_to_images(self, pdf_content_stream):
        try:
            images = convert_from_bytes(pdf_content_stream.read(), dpi=300)
            return images
        except Exception as e:
            return []
    
    def __process_pdf_and_detect_labels(self, pdf_content, confidence_threshold=0.85):
        try:
            # Convert the PDF content bytes into a readable stream
            pdf_content_stream = BytesIO(pdf_content)

            # Convert the first page of the PDF to an image
            images = self.convert_pdf_to_images(pdf_content_stream)

            if images:
                # Convert PIL image to a bytes-like object
                img_byte_array = BytesIO()
                images[0].save(img_byte_array, format='JPEG')
                img_byte_array.seek(0)
                # print(f"img arr: {img_byte_array}")

                # Load the YOLOv5 model
                model_path = pkg_resources.resource_filename('bankstatementextractor_final', 'models/best.pt')
                model = torch.hub.load('ultralytics/yolov5', 'custom', path=model_path)

                # Load the image from the bytes-like object
                img = Image.open(img_byte_array)

                # Perform inference
                results = model(img)

                # Get bounding box coordinates, confidence scores, and labels
                boxes = results.pred[0][:, :4].cpu().numpy()
                confidences = results.pred[0][:, 4].cpu().numpy()
                labels = results.pred[0][:, 5].cpu().numpy().astype(int)

                detected_labels = []

                # Filter detections based on confidence threshold and collect labels
                for box, confidence, label in zip(boxes, confidences, labels):
                    if confidence >= confidence_threshold:
                        label_name = model.names[label]
                        detected_labels.append(label_name)

                return detected_labels
            else:
                return []
        except Exception as e:
            return []
            
            
    def extract(self, pdf_bytes):
        data=None
        res={}
        # Check PDF metadata
        metadata_result, metadata_message = self.__check_pdf_metadata(pdf_bytes)
        if metadata_result:
            # Process PDF and detect labels
            detected_labels = self.__process_pdf_and_detect_labels(pdf_bytes)
            if detected_labels and len(detected_labels)==1:
                print("Detected Labels:", detected_labels)
                banks = Banks()
                label = detected_labels[0]
                if label in const.bank_labels:
                    data=getattr(banks, label)(pdf_bytes)
                else:
                    print(f"No function found for label: {label}")
            else:
                print("No labels detected.")
        else:
            print("Metadata Check Failed:", metadata_message)

        if data:
            res=Incom_expense().income_detection(data)
        return res
        
